{% extends "base.html" %}
{% block content %}
<div class="jumbotron">
    <div class="container">
        <h1>FASTR Web Client</h1>
        <p>Welcome to the fastr web client! From here you can manage and view everything FASTR has to offer!</p>
        <p><a class="btn btn-primary btn-lg" href="#" role="button">Learn more &raquo;</a></p>
    </div>
</div>

<div class="container">
    <!-- row of 3 columns -->
    <div class="col-md-4">
        <h2>Websocket client</h2>
        <p>A websocket client tester. This tool will help you debugging and developing the websocket message passing between the fastr-server and client.</p>
        <p><a class="btn btn-default" href="{{ url_for('websocket_client') }}" role="button">View details &raquo;</a></p>
    </div>
    <div class="col-md-4">
        <h2>Tools</h2>
        <p>Browse all tools that are installed in this instance of fastr. You can get a convenient overview over which tools are installed as well as which versions are installed. The definition of the tool is also rendered here for your convenience.</p>
        <p><a class="btn btn-default" href="{{ url_for('tool') }}" role="button">View details &raquo;</a></p>
    </div>
    <div class="col-md-4">
        <h2>Networks</h2>
        <p>View existing networks and inspect runs.</p>
        <p><a class="btn btn-default" href="{{ url_for('networks') }}" role="button">View details &raquo;</a></p>
    </div>
    <div class="col-md-4">
        <h2>Docs</h2>
        <p>A locally compiled version of the docs. Make sure you make a symlink from web/static/doc to doc/_build/html</p>
        <p><a class="btn btn-default" href="{{ url_for('doc') }}" role="button">View details &raquo;</a></p>
    </div>
</div>

{% endblock %}